r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Iam
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional
from twilio.base import deserialize, values
from twilio.base.api_response import ApiResponse
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class OAuthAppInstance(InstanceResource):

    class IamV1AccountVendorOauthAppCreateRequest(object):
        """
        :ivar type:
        :ivar friendly_name:
        :ivar owner_sid:
        :ivar description:
        :ivar client_sid:
        :ivar policy:
        :ivar access_token_ttl:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.friendly_name: Optional[str] = payload.get("friendly_name")
            self.owner_sid: Optional[str] = payload.get("owner_sid")
            self.description: Optional[str] = payload.get("description")
            self.client_sid: Optional[str] = payload.get("client_sid")
            self.policy: Optional[
                OAuthAppList.IamV1OrganizationVendoroauthappPolicy
            ] = payload.get("policy")
            self.access_token_ttl: Optional[int] = payload.get("access_token_ttl")

        def to_dict(self):
            return {
                "type": self.type,
                "friendly_name": self.friendly_name,
                "owner_sid": self.owner_sid,
                "description": self.description,
                "client_sid": self.client_sid,
                "policy": self.policy.to_dict() if self.policy is not None else None,
                "access_token_ttl": self.access_token_ttl,
            }

    class IamV1AccountVendorOauthAppUpdateRequest(object):
        """
        :ivar type:
        :ivar friendly_name:
        :ivar description:
        :ivar policy:
        :ivar access_token_ttl:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.friendly_name: Optional[str] = payload.get("friendly_name")
            self.description: Optional[str] = payload.get("description")
            self.policy: Optional[
                OAuthAppList.IamV1OrganizationVendorOauthAppUpdateRequestPolicy
            ] = payload.get("policy")
            self.access_token_ttl: Optional[int] = payload.get("access_token_ttl")

        def to_dict(self):
            return {
                "type": self.type,
                "friendly_name": self.friendly_name,
                "description": self.description,
                "policy": self.policy.to_dict() if self.policy is not None else None,
                "access_token_ttl": self.access_token_ttl,
            }

    class IamV1OrganizationVendorOauthAppUpdateRequestPolicy(object):
        """
        :ivar allow: Set of permissions explicitly allowed
        :ivar deny: Set of permissions explicitly denied
        """

        def __init__(self, payload: Dict[str, Any]):

            self.allow: Optional[List[str]] = payload.get("allow")
            self.deny: Optional[List[str]] = payload.get("deny")

        def to_dict(self):
            return {
                "allow": self.allow,
                "deny": self.deny,
            }

    class IamV1OrganizationVendoroauthappPolicy(object):
        """
        :ivar allow: Set of permissions explicitly allowed
        :ivar deny: Set of permissions explicitly denied
        """

        def __init__(self, payload: Dict[str, Any]):

            self.allow: Optional[List[str]] = payload.get("allow")
            self.deny: Optional[List[str]] = payload.get("deny")

        def to_dict(self):
            return {
                "allow": self.allow,
                "deny": self.deny,
            }

    """
    :ivar type: 
    :ivar sid: 
    :ivar friendly_name: 
    :ivar description: 
    :ivar date_created: 
    :ivar created_by: 
    :ivar secret: 
    :ivar status: 
    :ivar policy: 
    :ivar access_token_ttl: 
    :ivar code: Twilio-specific error code
    :ivar message: Error message
    :ivar more_info: Link to Error Code References
    """

    def __init__(
        self, version: Version, payload: Dict[str, Any], sid: Optional[str] = None
    ):
        super().__init__(version)

        self.type: Optional[str] = payload.get("type")
        self.sid: Optional[str] = payload.get("sid")
        self.friendly_name: Optional[str] = payload.get("friendly_name")
        self.description: Optional[str] = payload.get("description")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.created_by: Optional[str] = payload.get("created_by")
        self.secret: Optional[str] = payload.get("secret")
        self.status: Optional[str] = payload.get("status")
        self.policy: Optional[str] = payload.get("policy")
        self.access_token_ttl: Optional[int] = deserialize.integer(
            payload.get("access_token_ttl")
        )
        self.code: Optional[int] = payload.get("code")
        self.message: Optional[str] = payload.get("message")
        self.more_info: Optional[str] = payload.get("more_info")

        self._solution = {
            "sid": sid or self.sid,
        }
        self._context: Optional[OAuthAppContext] = None

    @property
    def _proxy(self) -> "OAuthAppContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: OAuthAppContext for this OAuthAppInstance
        """
        if self._context is None:
            self._context = OAuthAppContext(
                self._version,
                sid=self._solution["sid"],
            )
        return self._context

    def delete(self) -> bool:
        """
        Deletes the OAuthAppInstance


        :returns: True if delete succeeds, False otherwise
        """
        return self._proxy.delete()

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the OAuthAppInstance


        :returns: True if delete succeeds, False otherwise
        """
        return await self._proxy.delete_async()

    def delete_with_http_info(self) -> ApiResponse:
        """
        Deletes the OAuthAppInstance with HTTP info


        :returns: ApiResponse with success boolean, status code, and headers
        """
        return self._proxy.delete_with_http_info()

    async def delete_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine that deletes the OAuthAppInstance with HTTP info


        :returns: ApiResponse with success boolean, status code, and headers
        """
        return await self._proxy.delete_with_http_info_async()

    def update(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> "OAuthAppInstance":
        """
        Update the OAuthAppInstance

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: The updated OAuthAppInstance
        """
        return self._proxy.update(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request,
        )

    async def update_async(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> "OAuthAppInstance":
        """
        Asynchronous coroutine to update the OAuthAppInstance

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: The updated OAuthAppInstance
        """
        return await self._proxy.update_async(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request,
        )

    def update_with_http_info(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> ApiResponse:
        """
        Update the OAuthAppInstance with HTTP info

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        return self._proxy.update_with_http_info(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request,
        )

    async def update_with_http_info_async(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> ApiResponse:
        """
        Asynchronous coroutine to update the OAuthAppInstance with HTTP info

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        return await self._proxy.update_with_http_info_async(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Iam.V1.OAuthAppInstance {}>".format(context)


class OAuthAppContext(InstanceContext):

    class IamV1AccountVendorOauthAppCreateRequest(object):
        """
        :ivar type:
        :ivar friendly_name:
        :ivar owner_sid:
        :ivar description:
        :ivar client_sid:
        :ivar policy:
        :ivar access_token_ttl:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.friendly_name: Optional[str] = payload.get("friendly_name")
            self.owner_sid: Optional[str] = payload.get("owner_sid")
            self.description: Optional[str] = payload.get("description")
            self.client_sid: Optional[str] = payload.get("client_sid")
            self.policy: Optional[
                OAuthAppList.IamV1OrganizationVendoroauthappPolicy
            ] = payload.get("policy")
            self.access_token_ttl: Optional[int] = payload.get("access_token_ttl")

        def to_dict(self):
            return {
                "type": self.type,
                "friendly_name": self.friendly_name,
                "owner_sid": self.owner_sid,
                "description": self.description,
                "client_sid": self.client_sid,
                "policy": self.policy.to_dict() if self.policy is not None else None,
                "access_token_ttl": self.access_token_ttl,
            }

    class IamV1AccountVendorOauthAppUpdateRequest(object):
        """
        :ivar type:
        :ivar friendly_name:
        :ivar description:
        :ivar policy:
        :ivar access_token_ttl:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.friendly_name: Optional[str] = payload.get("friendly_name")
            self.description: Optional[str] = payload.get("description")
            self.policy: Optional[
                OAuthAppList.IamV1OrganizationVendorOauthAppUpdateRequestPolicy
            ] = payload.get("policy")
            self.access_token_ttl: Optional[int] = payload.get("access_token_ttl")

        def to_dict(self):
            return {
                "type": self.type,
                "friendly_name": self.friendly_name,
                "description": self.description,
                "policy": self.policy.to_dict() if self.policy is not None else None,
                "access_token_ttl": self.access_token_ttl,
            }

    class IamV1OrganizationVendorOauthAppUpdateRequestPolicy(object):
        """
        :ivar allow: Set of permissions explicitly allowed
        :ivar deny: Set of permissions explicitly denied
        """

        def __init__(self, payload: Dict[str, Any]):

            self.allow: Optional[List[str]] = payload.get("allow")
            self.deny: Optional[List[str]] = payload.get("deny")

        def to_dict(self):
            return {
                "allow": self.allow,
                "deny": self.deny,
            }

    class IamV1OrganizationVendoroauthappPolicy(object):
        """
        :ivar allow: Set of permissions explicitly allowed
        :ivar deny: Set of permissions explicitly denied
        """

        def __init__(self, payload: Dict[str, Any]):

            self.allow: Optional[List[str]] = payload.get("allow")
            self.deny: Optional[List[str]] = payload.get("deny")

        def to_dict(self):
            return {
                "allow": self.allow,
                "deny": self.deny,
            }

    def __init__(self, version: Version, sid: str):
        """
        Initialize the OAuthAppContext

        :param version: Version that contains the resource
        :param sid: Unique ID (sid) of the OAuth app
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "sid": sid,
        }
        self._uri = "/Account/OAuthApps/{sid}".format(**self._solution)

    def _delete(self) -> tuple:
        """
        Internal helper for delete operation

        Returns:
            tuple: (success_boolean, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return self._version.delete_with_response_info(
            method="DELETE", uri=self._uri, headers=headers
        )

    def delete(self) -> bool:
        """
        Deletes the OAuthAppInstance


        :returns: True if delete succeeds, False otherwise
        """
        success, _, _ = self._delete()
        return success

    def delete_with_http_info(self) -> ApiResponse:
        """
        Deletes the OAuthAppInstance and return response metadata


        :returns: ApiResponse with success boolean, status code, and headers
        """
        success, status_code, headers = self._delete()
        return ApiResponse(data=success, status_code=status_code, headers=headers)

    async def _delete_async(self) -> tuple:
        """
        Internal async helper for delete operation

        Returns:
            tuple: (success_boolean, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return await self._version.delete_with_response_info_async(
            method="DELETE", uri=self._uri, headers=headers
        )

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the OAuthAppInstance


        :returns: True if delete succeeds, False otherwise
        """
        success, _, _ = await self._delete_async()
        return success

    async def delete_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine that deletes the OAuthAppInstance and return response metadata


        :returns: ApiResponse with success boolean, status code, and headers
        """
        success, status_code, headers = await self._delete_async()
        return ApiResponse(data=success, status_code=status_code, headers=headers)

    def _update(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> tuple:
        """
        Internal helper for update operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = iam_v1_account_vendor_oauth_app_update_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return self._version.update_with_response_info(
            method="PUT", uri=self._uri, data=data, headers=headers
        )

    def update(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> OAuthAppInstance:
        """
        Update the OAuthAppInstance

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: The updated OAuthAppInstance
        """
        payload, _, _ = self._update(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request
        )
        return OAuthAppInstance(self._version, payload, sid=self._solution["sid"])

    def update_with_http_info(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> ApiResponse:
        """
        Update the OAuthAppInstance and return response metadata

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._update(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request
        )
        instance = OAuthAppInstance(self._version, payload, sid=self._solution["sid"])
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _update_async(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> tuple:
        """
        Internal async helper for update operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = iam_v1_account_vendor_oauth_app_update_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return await self._version.update_with_response_info_async(
            method="PUT", uri=self._uri, data=data, headers=headers
        )

    async def update_async(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> OAuthAppInstance:
        """
        Asynchronous coroutine to update the OAuthAppInstance

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: The updated OAuthAppInstance
        """
        payload, _, _ = await self._update_async(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request
        )
        return OAuthAppInstance(self._version, payload, sid=self._solution["sid"])

    async def update_with_http_info_async(
        self,
        iam_v1_account_vendor_oauth_app_update_request: IamV1AccountVendorOauthAppUpdateRequest,
    ) -> ApiResponse:
        """
        Asynchronous coroutine to update the OAuthAppInstance and return response metadata

        :param iam_v1_account_vendor_oauth_app_update_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._update_async(
            iam_v1_account_vendor_oauth_app_update_request=iam_v1_account_vendor_oauth_app_update_request
        )
        instance = OAuthAppInstance(self._version, payload, sid=self._solution["sid"])
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Iam.V1.OAuthAppContext {}>".format(context)


class OAuthAppList(ListResource):

    class IamV1AccountVendorOauthAppCreateRequest(object):
        """
        :ivar type:
        :ivar friendly_name:
        :ivar owner_sid:
        :ivar description:
        :ivar client_sid:
        :ivar policy:
        :ivar access_token_ttl:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.friendly_name: Optional[str] = payload.get("friendly_name")
            self.owner_sid: Optional[str] = payload.get("owner_sid")
            self.description: Optional[str] = payload.get("description")
            self.client_sid: Optional[str] = payload.get("client_sid")
            self.policy: Optional[
                OAuthAppList.IamV1OrganizationVendoroauthappPolicy
            ] = payload.get("policy")
            self.access_token_ttl: Optional[int] = payload.get("access_token_ttl")

        def to_dict(self):
            return {
                "type": self.type,
                "friendly_name": self.friendly_name,
                "owner_sid": self.owner_sid,
                "description": self.description,
                "client_sid": self.client_sid,
                "policy": self.policy.to_dict() if self.policy is not None else None,
                "access_token_ttl": self.access_token_ttl,
            }

    class IamV1AccountVendorOauthAppUpdateRequest(object):
        """
        :ivar type:
        :ivar friendly_name:
        :ivar description:
        :ivar policy:
        :ivar access_token_ttl:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.friendly_name: Optional[str] = payload.get("friendly_name")
            self.description: Optional[str] = payload.get("description")
            self.policy: Optional[
                OAuthAppList.IamV1OrganizationVendorOauthAppUpdateRequestPolicy
            ] = payload.get("policy")
            self.access_token_ttl: Optional[int] = payload.get("access_token_ttl")

        def to_dict(self):
            return {
                "type": self.type,
                "friendly_name": self.friendly_name,
                "description": self.description,
                "policy": self.policy.to_dict() if self.policy is not None else None,
                "access_token_ttl": self.access_token_ttl,
            }

    class IamV1OrganizationVendorOauthAppUpdateRequestPolicy(object):
        """
        :ivar allow: Set of permissions explicitly allowed
        :ivar deny: Set of permissions explicitly denied
        """

        def __init__(self, payload: Dict[str, Any]):

            self.allow: Optional[List[str]] = payload.get("allow")
            self.deny: Optional[List[str]] = payload.get("deny")

        def to_dict(self):
            return {
                "allow": self.allow,
                "deny": self.deny,
            }

    class IamV1OrganizationVendoroauthappPolicy(object):
        """
        :ivar allow: Set of permissions explicitly allowed
        :ivar deny: Set of permissions explicitly denied
        """

        def __init__(self, payload: Dict[str, Any]):

            self.allow: Optional[List[str]] = payload.get("allow")
            self.deny: Optional[List[str]] = payload.get("deny")

        def to_dict(self):
            return {
                "allow": self.allow,
                "deny": self.deny,
            }

    def __init__(self, version: Version):
        """
        Initialize the OAuthAppList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/Account/OAuthApps"

    def _create(
        self,
        iam_v1_account_vendor_oauth_app_create_request: IamV1AccountVendorOauthAppCreateRequest,
    ) -> tuple:
        """
        Internal helper for create operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = iam_v1_account_vendor_oauth_app_create_request.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return self._version.create_with_response_info(
            method="POST", uri=self._uri, data=data, headers=headers
        )

    def create(
        self,
        iam_v1_account_vendor_oauth_app_create_request: IamV1AccountVendorOauthAppCreateRequest,
    ) -> OAuthAppInstance:
        """
        Create the OAuthAppInstance

        :param iam_v1_account_vendor_oauth_app_create_request:

        :returns: The created OAuthAppInstance
        """
        payload, _, _ = self._create(
            iam_v1_account_vendor_oauth_app_create_request=iam_v1_account_vendor_oauth_app_create_request
        )
        return OAuthAppInstance(self._version, payload)

    def create_with_http_info(
        self,
        iam_v1_account_vendor_oauth_app_create_request: IamV1AccountVendorOauthAppCreateRequest,
    ) -> ApiResponse:
        """
        Create the OAuthAppInstance and return response metadata

        :param iam_v1_account_vendor_oauth_app_create_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._create(
            iam_v1_account_vendor_oauth_app_create_request=iam_v1_account_vendor_oauth_app_create_request
        )
        instance = OAuthAppInstance(self._version, payload)
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _create_async(
        self,
        iam_v1_account_vendor_oauth_app_create_request: IamV1AccountVendorOauthAppCreateRequest,
    ) -> tuple:
        """
        Internal async helper for create operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = iam_v1_account_vendor_oauth_app_create_request.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return await self._version.create_with_response_info_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

    async def create_async(
        self,
        iam_v1_account_vendor_oauth_app_create_request: IamV1AccountVendorOauthAppCreateRequest,
    ) -> OAuthAppInstance:
        """
        Asynchronously create the OAuthAppInstance

        :param iam_v1_account_vendor_oauth_app_create_request:

        :returns: The created OAuthAppInstance
        """
        payload, _, _ = await self._create_async(
            iam_v1_account_vendor_oauth_app_create_request=iam_v1_account_vendor_oauth_app_create_request
        )
        return OAuthAppInstance(self._version, payload)

    async def create_with_http_info_async(
        self,
        iam_v1_account_vendor_oauth_app_create_request: IamV1AccountVendorOauthAppCreateRequest,
    ) -> ApiResponse:
        """
        Asynchronously create the OAuthAppInstance and return response metadata

        :param iam_v1_account_vendor_oauth_app_create_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._create_async(
            iam_v1_account_vendor_oauth_app_create_request=iam_v1_account_vendor_oauth_app_create_request
        )
        instance = OAuthAppInstance(self._version, payload)
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def get(self, sid: str) -> OAuthAppContext:
        """
        Constructs a OAuthAppContext

        :param sid: Unique ID (sid) of the OAuth app
        """
        return OAuthAppContext(self._version, sid=sid)

    def __call__(self, sid: str) -> OAuthAppContext:
        """
        Constructs a OAuthAppContext

        :param sid: Unique ID (sid) of the OAuth app
        """
        return OAuthAppContext(self._version, sid=sid)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Iam.V1.OAuthAppList>"



from typing import Optional, Any, Dict

from twilio.base.exceptions import TwilioException
from twilio.base.page import Page


class TokenPagination(Page):
    """
    Represents a page of records using token-based pagination.

    Token-based pagination uses next_token and previous_token instead of
    page numbers to navigate through results.

    Example expected response format with token metadata:
    {
      "meta": {
        "key": "items",
        "pageSize": 50,
        "nextToken": "abc123",
        "previousToken": "xyz789"
      },
      "items": [
        { "id": 1, "name": "Item 1" },
        { "id": 2, "name": "Item 2" }
      ]
    }
    """

    def __init__(
        self,
        version,
        response,
        uri: str,
        params: Optional[Dict[str, Any]] = None,
        solution: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(version, response, solution)
        self._uri = uri
        self._params = params if params is not None else {}

    @property
    def key(self) -> Optional[str]:
        """
        :return str: Returns the key that identifies the collection in the response.
        """
        if "meta" in self._payload and "key" in self._payload["meta"]:
            return self._payload["meta"]["key"]
        return None

    @property
    def page_size(self) -> Optional[int]:
        """
        :return int: Returns the page size or None if doesn't exist.
        """
        if "meta" in self._payload and "pageSize" in self._payload["meta"]:
            return self._payload["meta"]["pageSize"]
        return None

    @property
    def next_token(self) -> Optional[str]:
        """
        :return str: Returns the next_token for pagination or None if doesn't exist.
        """
        if "meta" in self._payload and "nextToken" in self._payload["meta"]:
            return self._payload["meta"]["nextToken"]
        return None

    @property
    def previous_token(self) -> Optional[str]:
        """
        :return str: Returns the previous_token for pagination or None if doesn't exist.
        """
        if "meta" in self._payload and "previousToken" in self._payload["meta"]:
            return self._payload["meta"]["previousToken"]
        return None

    def _get_page(self, token: Optional[str]) -> Optional["TokenPagination"]:
        """
        Internal helper to fetch a page using a given token.

        :param token: The pagination token to use.
        :return: The page or None if no token exists.
        """
        if not token:
            return None

        if not self._uri:
            raise TwilioException("URI must be provided for token pagination")

        self._params["pageToken"] = token

        response = self._version.page(method="GET", uri=self._uri, params=self._params)
        cls = type(self)
        return cls(self._version, response, self._uri, self._params, self._solution)

    async def _get_page_async(
        self, token: Optional[str]
    ) -> Optional["TokenPagination"]:
        """
        Internal async helper to fetch a page using a given token.

        :param token: The pagination token to use.
        :return: The page or None if no token exists.
        """
        if not token:
            return None

        if not self._uri:
            raise TwilioException("URI must be provided for token pagination")

        # Construct full URL with pageToken parameter
        self._params["pageToken"] = token
        response = self._version.page(method="GET", uri=self._uri, params=self._params)
        cls = type(self)
        return cls(self._version, response, self._uri, self._params, self._solution)

    def next_page(self) -> Optional["TokenPagination"]:
        """
        Return the next page using token-based pagination.
        Makes a request to the same URI with pageToken set to nextToken.

        :return: The next page or None if no next token exists.
        """
        return self._get_page(self.next_token)

    async def next_page_async(self) -> Optional["TokenPagination"]:
        """
        Asynchronously return the next page using token-based pagination.
        Makes a request to the same URI with pageToken set to nextToken.

        :return: The next page or None if no next token exists.
        """
        return await self._get_page_async(self.next_token)

    def previous_page(self) -> Optional["TokenPagination"]:
        """
        Return the previous page using token-based pagination.
        Makes a request to the same URI with pageToken set to previousToken.

        :return: The previous page or None if no previous token exists.
        """
        return self._get_page(self.previous_token)

    async def previous_page_async(self) -> Optional["TokenPagination"]:
        """
        Asynchronously return the previous page using token-based pagination.
        Makes a request to the same URI with pageToken set to previousToken.

        :return: The previous page or None if no previous token exists.
        """
        return await self._get_page_async(self.previous_token)

    def __repr__(self) -> str:
        return "<TokenPagination>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Pricing
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import values
from twilio.base.api_response import ApiResponse
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class CountryInstance(InstanceResource):
    """
    :ivar country: The name of the country.
    :ivar iso_country: The [ISO country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2).
    :ivar terminating_prefix_prices: The list of [TerminatingPrefixPrice](https://www.twilio.com/docs/voice/pricing#outbound-prefix-price-with-origin) records.
    :ivar originating_call_prices: The list of [OriginatingCallPrice](https://www.twilio.com/docs/voice/pricing#inbound-call-price) records.
    :ivar price_unit: The currency in which prices are measured, specified in [ISO 4127](https://www.iso.org/iso/home/standards/currency_codes.htm) format (e.g. `usd`, `eur`, `jpy`).
    :ivar url: The absolute URL of the resource.
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        iso_country: Optional[str] = None,
    ):
        super().__init__(version)

        self.country: Optional[str] = payload.get("country")
        self.iso_country: Optional[str] = payload.get("iso_country")
        self.terminating_prefix_prices: Optional[List[str]] = payload.get(
            "terminating_prefix_prices"
        )
        self.originating_call_prices: Optional[List[str]] = payload.get(
            "originating_call_prices"
        )
        self.price_unit: Optional[str] = payload.get("price_unit")
        self.url: Optional[str] = payload.get("url")

        self._solution = {
            "iso_country": iso_country or self.iso_country,
        }
        self._context: Optional[CountryContext] = None

    @property
    def _proxy(self) -> "CountryContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: CountryContext for this CountryInstance
        """
        if self._context is None:
            self._context = CountryContext(
                self._version,
                iso_country=self._solution["iso_country"],
            )
        return self._context

    def fetch(self) -> "CountryInstance":
        """
        Fetch the CountryInstance


        :returns: The fetched CountryInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "CountryInstance":
        """
        Asynchronous coroutine to fetch the CountryInstance


        :returns: The fetched CountryInstance
        """
        return await self._proxy.fetch_async()

    def fetch_with_http_info(self) -> ApiResponse:
        """
        Fetch the CountryInstance with HTTP info


        :returns: ApiResponse with instance, status code, and headers
        """
        return self._proxy.fetch_with_http_info()

    async def fetch_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine to fetch the CountryInstance with HTTP info


        :returns: ApiResponse with instance, status code, and headers
        """
        return await self._proxy.fetch_with_http_info_async()

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Pricing.V2.CountryInstance {}>".format(context)


class CountryContext(InstanceContext):

    def __init__(self, version: Version, iso_country: str):
        """
        Initialize the CountryContext

        :param version: Version that contains the resource
        :param iso_country: The [ISO country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) of the origin-based voice pricing information to fetch.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "iso_country": iso_country,
        }
        self._uri = "/Trunking/Countries/{iso_country}".format(**self._solution)

    def _fetch(self) -> tuple:
        """
        Internal helper for fetch operation

        Returns:
            tuple: (payload, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return self._version.fetch_with_response_info(
            method="GET", uri=self._uri, headers=headers
        )

    def fetch(self) -> CountryInstance:
        """
        Fetch the CountryInstance


        :returns: The fetched CountryInstance
        """
        payload, _, _ = self._fetch()
        return CountryInstance(
            self._version,
            payload,
            iso_country=self._solution["iso_country"],
        )

    def fetch_with_http_info(self) -> ApiResponse:
        """
        Fetch the CountryInstance and return response metadata


        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._fetch()
        instance = CountryInstance(
            self._version,
            payload,
            iso_country=self._solution["iso_country"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _fetch_async(self) -> tuple:
        """
        Internal async helper for fetch operation

        Returns:
            tuple: (payload, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return await self._version.fetch_with_response_info_async(
            method="GET", uri=self._uri, headers=headers
        )

    async def fetch_async(self) -> CountryInstance:
        """
        Asynchronous coroutine to fetch the CountryInstance


        :returns: The fetched CountryInstance
        """
        payload, _, _ = await self._fetch_async()
        return CountryInstance(
            self._version,
            payload,
            iso_country=self._solution["iso_country"],
        )

    async def fetch_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine to fetch the CountryInstance and return response metadata


        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._fetch_async()
        instance = CountryInstance(
            self._version,
            payload,
            iso_country=self._solution["iso_country"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Pricing.V2.CountryContext {}>".format(context)


class CountryPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> CountryInstance:
        """
        Build an instance of CountryInstance

        :param payload: Payload response from the API
        """
        return CountryInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Pricing.V2.CountryPage>"


class CountryList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the CountryList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/Trunking/Countries"

    def stream(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[CountryInstance]:
        """
        Streams CountryInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[CountryInstance]:
        """
        Asynchronously streams CountryInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(page_size=limits["page_size"])

        return self._version.stream_async(page, limits["limit"])

    def stream_with_http_info(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> tuple:
        """
        Streams CountryInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: tuple of (generator, status_code, headers) where generator yields instances
        """
        limits = self._version.read_limits(limit, page_size)
        page_response = self.page_with_http_info(page_size=limits["page_size"])

        generator = self._version.stream(page_response.data, limits["limit"])
        return (generator, page_response.status_code, page_response.headers)

    async def stream_with_http_info_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> tuple:
        """
        Asynchronously streams CountryInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: tuple of (generator, status_code, headers) where generator yields instances
        """
        limits = self._version.read_limits(limit, page_size)
        page_response = await self.page_with_http_info_async(
            page_size=limits["page_size"]
        )

        generator = self._version.stream_async(page_response.data, limits["limit"])
        return (generator, page_response.status_code, page_response.headers)

    def list(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[CountryInstance]:
        """
        Lists CountryInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[CountryInstance]:
        """
        Asynchronously lists CountryInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                limit=limit,
                page_size=page_size,
            )
        ]

    def list_with_http_info(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ApiResponse:
        """
        Lists CountryInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: ApiResponse with list of instances, status code, and headers
        """
        generator, status_code, headers = self.stream_with_http_info(
            limit=limit,
            page_size=page_size,
        )
        items = list(generator)
        return ApiResponse(data=items, status_code=status_code, headers=headers)

    async def list_with_http_info_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ApiResponse:
        """
        Asynchronously lists CountryInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: ApiResponse with list of instances, status code, and headers
        """
        generator, status_code, headers = await self.stream_with_http_info_async(
            limit=limit,
            page_size=page_size,
        )
        items = [record async for record in generator]
        return ApiResponse(data=items, status_code=status_code, headers=headers)

    def page(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> CountryPage:
        """
        Retrieve a single page of CountryInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of CountryInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return CountryPage(self._version, response)

    async def page_async(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> CountryPage:
        """
        Asynchronously retrieve a single page of CountryInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of CountryInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return CountryPage(self._version, response)

    def page_with_http_info(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> ApiResponse:
        """
        Retrieve a single page with response metadata


        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: ApiResponse with CountryPage, status code, and headers
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response, status_code, response_headers = self._version.page_with_response_info(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        page = CountryPage(self._version, response)
        return ApiResponse(data=page, status_code=status_code, headers=response_headers)

    async def page_with_http_info_async(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> ApiResponse:
        """
        Asynchronously retrieve a single page with response metadata


        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: ApiResponse with CountryPage, status code, and headers
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response, status_code, response_headers = (
            await self._version.page_with_response_info_async(
                method="GET", uri=self._uri, params=data, headers=headers
            )
        )
        page = CountryPage(self._version, response)
        return ApiResponse(data=page, status_code=status_code, headers=response_headers)

    def get_page(self, target_url: str) -> CountryPage:
        """
        Retrieve a specific page of CountryInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of CountryInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return CountryPage(self._version, response)

    async def get_page_async(self, target_url: str) -> CountryPage:
        """
        Asynchronously retrieve a specific page of CountryInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of CountryInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return CountryPage(self._version, response)

    def get(self, iso_country: str) -> CountryContext:
        """
        Constructs a CountryContext

        :param iso_country: The [ISO country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) of the origin-based voice pricing information to fetch.
        """
        return CountryContext(self._version, iso_country=iso_country)

    def __call__(self, iso_country: str) -> CountryContext:
        """
        Constructs a CountryContext

        :param iso_country: The [ISO country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) of the origin-based voice pricing information to fetch.
        """
        return CountryContext(self._version, iso_country=iso_country)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Pricing.V2.CountryList>"



# -*- coding: utf-8 -*-
import sys
from typing import Optional, List, Dict


class TwilioException(Exception):
    pass


class TwilioRestException(TwilioException):
    """A generic 400 or 500 level exception from the Twilio API

    :param int status: the HTTP status that was returned for the exception
    :param str uri: The URI that caused the exception
    :param str msg: A human-readable message for the error
    :param int|None code: A Twilio-specific error code for the error. This is
         not available for all errors.
    :param method: The HTTP method used to make the request
    :param details: Additional error details returned for the exception
    """

    def __init__(
        self,
        status: int,
        uri: str,
        msg: str = "",
        code: Optional[int] = None,
        method: str = "GET",
        details: Optional[object] = None,
    ):
        self.uri = uri
        self.status = status
        self.msg = msg
        self.code = code
        self.method = method
        self.details = details

    def __str__(self) -> str:
        """Try to pretty-print the exception, if this is going on screen."""

        def red(words: str) -> str:
            return "\033[31m\033[49m%s\033[0m" % words

        def white(words: str) -> str:
            return "\033[37m\033[49m%s\033[0m" % words

        def blue(words: str) -> str:
            return "\033[34m\033[49m%s\033[0m" % words

        def teal(words: str) -> str:
            return "\033[36m\033[49m%s\033[0m" % words

        def get_uri(code: int) -> str:
            return "https://www.twilio.com/docs/errors/{0}".format(code)

        # If it makes sense to print a human readable error message, try to
        # do it. The one problem is that someone might catch this error and
        # try to display the message from it to an end user.
        if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
            msg = (
                "\n{red_error} {request_was}\n\n{http_line}"
                "\n\n{twilio_returned}\n\n{message}\n".format(
                    red_error=red("HTTP Error"),
                    request_was=white("Your request was:"),
                    http_line=teal("%s %s" % (self.method, self.uri)),
                    twilio_returned=white("Twilio returned the following information:"),
                    message=blue(str(self.msg)),
                )
            )
            if self.code:
                msg = "".join(
                    [
                        msg,
                        "\n{more_info}\n\n{uri}\n\n".format(
                            more_info=white("More information may be available here:"),
                            uri=blue(get_uri(self.code)),
                        ),
                    ]
                )
            return msg
        else:
            return "HTTP {0} error: {1}".format(self.status, self.msg)


class TwilioServiceException(TwilioException):
    """An RFC-9457 compliant exception from a Twilio service

    This exception follows the Problem Details for HTTP APIs specification
    (RFC-9457). See https://www.rfc-editor.org/rfc/rfc9457.html

    :param str type: A URI reference that identifies the problem type
    :param str title: A short, human-readable summary of the problem type
    :param int status: The HTTP status code for this occurrence of the problem
    :param int code: Twilio-specific error code
    :param str|None detail: A human-readable explanation specific to this occurrence
    :param str|None instance: A URI reference identifying the specific occurrence
    :param list|None errors: Array of validation errors with detail and pointer fields
    :param str method: The HTTP method used to make the request
    :param str uri: The URI that caused the exception
    """

    def __init__(
        self,
        type_uri: str,
        title: str,
        status: int,
        code: int,
        detail: Optional[str] = None,
        instance: Optional[str] = None,
        errors: Optional[List[Dict[str, str]]] = None,
        method: str = "GET",
        uri: str = "",
    ):
        self.type = type_uri
        self.title = title
        self.status = status
        self.code = code
        self.detail = detail
        self.instance = instance
        self.errors = errors or []
        self.method = method
        self.uri = uri

    def __str__(self) -> str:
        """Pretty-print the exception for terminal output."""

        def red(words: str) -> str:
            return "\033[31m\033[49m%s\033[0m" % words

        def white(words: str) -> str:
            return "\033[37m\033[49m%s\033[0m" % words

        def blue(words: str) -> str:
            return "\033[34m\033[49m%s\033[0m" % words

        def teal(words: str) -> str:
            return "\033[36m\033[49m%s\033[0m" % words

        def yellow(words: str) -> str:
            return "\033[33m\033[49m%s\033[0m" % words

        # Check if we're in a TTY for colored output
        if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
            msg_parts = [
                "\n{red_error} {request_was}\n\n{http_line}\n\n{twilio_returned}\n".format(
                    red_error=red("HTTP Error"),
                    request_was=white("Your request was:"),
                    http_line=(
                        teal("%s %s" % (self.method, self.uri))
                        if self.uri
                        else teal("(no URI)")
                    ),
                    twilio_returned=white("Twilio returned the following information:"),
                )
            ]

            # Title and detail
            msg_parts.append(
                "\n{title_label}: {title}\n".format(
                    title_label=white("Title"),
                    title=blue(self.title),
                )
            )

            if self.detail:
                msg_parts.append(
                    "{detail_label}: {detail}\n".format(
                        detail_label=white("Detail"),
                        detail=blue(self.detail),
                    )
                )

            # Code and status
            msg_parts.append(
                "{code_label}: {code} | {status_label}: {status}\n".format(
                    code_label=white("Error Code"),
                    code=blue(str(self.code)),
                    status_label=white("Status"),
                    status=blue(str(self.status)),
                )
            )

            # Validation errors if present
            if self.errors:
                msg_parts.append(
                    "\n{validation_label}:\n".format(
                        validation_label=white("Validation Errors"),
                    )
                )
                for error in self.errors:
                    msg_parts.append(
                        "  {pointer}: {detail}\n".format(
                            pointer=yellow(error.get("pointer", "(unknown field)")),
                            detail=error.get("detail", "(no detail)"),
                        )
                    )

            # Documentation link
            msg_parts.append(
                "\n{more_info}\n\n{uri}\n\n".format(
                    more_info=white("More information may be available here:"),
                    uri=blue(self.type),
                )
            )

            return "".join(msg_parts)
        else:
            # Plain text output for non-TTY
            msg = "HTTP {0} error: {1}".format(self.status, self.title)
            if self.detail:
                msg += " - {0}".format(self.detail)
            if self.errors:
                msg += " | Validation errors: {0}".format(
                    ", ".join(
                        "{0} ({1})".format(e.get("pointer", "?"), e.get("detail", "?"))
                        for e in self.errors
                    )
                )
            return msg



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Lookups
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, Optional, Union
from twilio.base import deserialize, values
from twilio.base.api_response import ApiResponse
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class LookupOverrideInstance(InstanceResource):

    class OverridesRequest(object):
        """
        :ivar line_type: The new line type to override the original line type
        :ivar reason: The reason for the override
        """

        def __init__(self, payload: Dict[str, Any]):

            self.line_type: Optional["LookupOverrideInstance.str"] = payload.get(
                "line_type"
            )
            self.reason: Optional[str] = payload.get("reason")

        def to_dict(self):
            return {
                "line_type": self.line_type,
                "reason": self.reason,
            }

    """
    :ivar phone_number: The phone number for which the override was created
    :ivar original_line_type: The original line type
    :ivar overridden_line_type: The new line type after the override
    :ivar override_reason: The reason for the override
    :ivar override_timestamp: 
    :ivar overridden_by_account_sid: The Account SID for the user who made the override
    :ivar code: Twilio-specific error code
    :ivar message: Error message
    :ivar more_info: Link to Error Code References
    :ivar status: HTTP response status code
    :ivar field: Limit of requests for the bucket
    :ivar limit: Limit of requests for the bucket
    :ivar bucket: Name of the bucket
    :ivar owner: Owner of the rule
    :ivar ttl: Time to live of the rule
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        field: Optional[str] = None,
        phone_number: Optional[str] = None,
    ):
        super().__init__(version)

        self.phone_number: Optional[str] = payload.get("phone_number")
        self.original_line_type: Optional[str] = payload.get("original_line_type")
        self.overridden_line_type: Optional[str] = payload.get("overridden_line_type")
        self.override_reason: Optional[str] = payload.get("override_reason")
        self.override_timestamp: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("override_timestamp")
        )
        self.overridden_by_account_sid: Optional[str] = payload.get(
            "overridden_by_account_sid"
        )
        self.code: Optional[int] = payload.get("code")
        self.message: Optional[str] = payload.get("message")
        self.more_info: Optional[str] = payload.get("more_info")
        self.status: Optional[int] = payload.get("status")
        self.field: Optional[str] = payload.get("field")
        self.limit: Optional[int] = payload.get("limit")
        self.bucket: Optional[str] = payload.get("bucket")
        self.owner: Optional[str] = payload.get("owner")
        self.ttl: Optional[int] = payload.get("ttl")

        self._solution = {
            "field": field or self.field,
            "phone_number": phone_number or self.phone_number,
        }
        self._context: Optional[LookupOverrideContext] = None

    @property
    def _proxy(self) -> "LookupOverrideContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: LookupOverrideContext for this LookupOverrideInstance
        """
        if self._context is None:
            self._context = LookupOverrideContext(
                self._version,
                field=self._solution["field"],
                phone_number=self._solution["phone_number"],
            )
        return self._context

    def create(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> "LookupOverrideInstance":
        """
        Create the LookupOverrideInstance

        :param overrides_request:

        :returns: The created LookupOverrideInstance
        """
        return self._proxy.create(
            overrides_request=overrides_request,
        )

    async def create_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> "LookupOverrideInstance":
        """
        Asynchronous coroutine to create the LookupOverrideInstance

        :param overrides_request:

        :returns: The created LookupOverrideInstance
        """
        return await self._proxy.create_async(
            overrides_request=overrides_request,
        )

    def create_with_http_info(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Create the LookupOverrideInstance with HTTP info

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        return self._proxy.create_with_http_info(
            overrides_request=overrides_request,
        )

    async def create_with_http_info_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Asynchronous coroutine to create the LookupOverrideInstance with HTTP info

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        return await self._proxy.create_with_http_info_async(
            overrides_request=overrides_request,
        )

    def delete(self) -> bool:
        """
        Deletes the LookupOverrideInstance


        :returns: True if delete succeeds, False otherwise
        """
        return self._proxy.delete()

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the LookupOverrideInstance


        :returns: True if delete succeeds, False otherwise
        """
        return await self._proxy.delete_async()

    def delete_with_http_info(self) -> ApiResponse:
        """
        Deletes the LookupOverrideInstance with HTTP info


        :returns: ApiResponse with success boolean, status code, and headers
        """
        return self._proxy.delete_with_http_info()

    async def delete_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine that deletes the LookupOverrideInstance with HTTP info


        :returns: ApiResponse with success boolean, status code, and headers
        """
        return await self._proxy.delete_with_http_info_async()

    def fetch(self) -> "LookupOverrideInstance":
        """
        Fetch the LookupOverrideInstance


        :returns: The fetched LookupOverrideInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "LookupOverrideInstance":
        """
        Asynchronous coroutine to fetch the LookupOverrideInstance


        :returns: The fetched LookupOverrideInstance
        """
        return await self._proxy.fetch_async()

    def fetch_with_http_info(self) -> ApiResponse:
        """
        Fetch the LookupOverrideInstance with HTTP info


        :returns: ApiResponse with instance, status code, and headers
        """
        return self._proxy.fetch_with_http_info()

    async def fetch_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine to fetch the LookupOverrideInstance with HTTP info


        :returns: ApiResponse with instance, status code, and headers
        """
        return await self._proxy.fetch_with_http_info_async()

    def update(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> "LookupOverrideInstance":
        """
        Update the LookupOverrideInstance

        :param overrides_request:

        :returns: The updated LookupOverrideInstance
        """
        return self._proxy.update(
            overrides_request=overrides_request,
        )

    async def update_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> "LookupOverrideInstance":
        """
        Asynchronous coroutine to update the LookupOverrideInstance

        :param overrides_request:

        :returns: The updated LookupOverrideInstance
        """
        return await self._proxy.update_async(
            overrides_request=overrides_request,
        )

    def update_with_http_info(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Update the LookupOverrideInstance with HTTP info

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        return self._proxy.update_with_http_info(
            overrides_request=overrides_request,
        )

    async def update_with_http_info_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Asynchronous coroutine to update the LookupOverrideInstance with HTTP info

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        return await self._proxy.update_with_http_info_async(
            overrides_request=overrides_request,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Lookups.V2.LookupOverrideInstance {}>".format(context)


class LookupOverrideContext(InstanceContext):

    class OverridesRequest(object):
        """
        :ivar line_type: The new line type to override the original line type
        :ivar reason: The reason for the override
        """

        def __init__(self, payload: Dict[str, Any]):

            self.line_type: Optional["LookupOverrideInstance.str"] = payload.get(
                "line_type"
            )
            self.reason: Optional[str] = payload.get("reason")

        def to_dict(self):
            return {
                "line_type": self.line_type,
                "reason": self.reason,
            }

    def __init__(self, version: Version, field: str, phone_number: str):
        """
        Initialize the LookupOverrideContext

        :param version: Version that contains the resource
        :param field:
        :param phone_number:
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "field": field,
            "phone_number": phone_number,
        }
        self._uri = "/PhoneNumbers/{phone_number}/Overrides/{field}".format(
            **self._solution
        )

    def _create(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> tuple:
        """
        Internal helper for create operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = overrides_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return self._version.create_with_response_info(
            method="POST", uri=self._uri, data=data, headers=headers
        )

    def create(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> LookupOverrideInstance:
        """
        Create the LookupOverrideInstance

        :param overrides_request:

        :returns: The created LookupOverrideInstance
        """
        payload, _, _ = self._create(overrides_request=overrides_request)
        return LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )

    def create_with_http_info(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Create the LookupOverrideInstance and return response metadata

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._create(
            overrides_request=overrides_request
        )
        instance = LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _create_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> tuple:
        """
        Internal async helper for create operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = overrides_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return await self._version.create_with_response_info_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

    async def create_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> LookupOverrideInstance:
        """
        Asynchronous coroutine to create the LookupOverrideInstance

        :param overrides_request:

        :returns: The created LookupOverrideInstance
        """
        payload, _, _ = await self._create_async(overrides_request=overrides_request)
        return LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )

    async def create_with_http_info_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Asynchronous coroutine to create the LookupOverrideInstance and return response metadata

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._create_async(
            overrides_request=overrides_request
        )
        instance = LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def _delete(self) -> tuple:
        """
        Internal helper for delete operation

        Returns:
            tuple: (success_boolean, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return self._version.delete_with_response_info(
            method="DELETE", uri=self._uri, headers=headers
        )

    def delete(self) -> bool:
        """
        Deletes the LookupOverrideInstance


        :returns: True if delete succeeds, False otherwise
        """
        success, _, _ = self._delete()
        return success

    def delete_with_http_info(self) -> ApiResponse:
        """
        Deletes the LookupOverrideInstance and return response metadata


        :returns: ApiResponse with success boolean, status code, and headers
        """
        success, status_code, headers = self._delete()
        return ApiResponse(data=success, status_code=status_code, headers=headers)

    async def _delete_async(self) -> tuple:
        """
        Internal async helper for delete operation

        Returns:
            tuple: (success_boolean, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return await self._version.delete_with_response_info_async(
            method="DELETE", uri=self._uri, headers=headers
        )

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the LookupOverrideInstance


        :returns: True if delete succeeds, False otherwise
        """
        success, _, _ = await self._delete_async()
        return success

    async def delete_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine that deletes the LookupOverrideInstance and return response metadata


        :returns: ApiResponse with success boolean, status code, and headers
        """
        success, status_code, headers = await self._delete_async()
        return ApiResponse(data=success, status_code=status_code, headers=headers)

    def _fetch(self) -> tuple:
        """
        Internal helper for fetch operation

        Returns:
            tuple: (payload, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return self._version.fetch_with_response_info(
            method="GET", uri=self._uri, headers=headers
        )

    def fetch(self) -> LookupOverrideInstance:
        """
        Fetch the LookupOverrideInstance


        :returns: The fetched LookupOverrideInstance
        """
        payload, _, _ = self._fetch()
        return LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )

    def fetch_with_http_info(self) -> ApiResponse:
        """
        Fetch the LookupOverrideInstance and return response metadata


        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._fetch()
        instance = LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _fetch_async(self) -> tuple:
        """
        Internal async helper for fetch operation

        Returns:
            tuple: (payload, status_code, headers)
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        return await self._version.fetch_with_response_info_async(
            method="GET", uri=self._uri, headers=headers
        )

    async def fetch_async(self) -> LookupOverrideInstance:
        """
        Asynchronous coroutine to fetch the LookupOverrideInstance


        :returns: The fetched LookupOverrideInstance
        """
        payload, _, _ = await self._fetch_async()
        return LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )

    async def fetch_with_http_info_async(self) -> ApiResponse:
        """
        Asynchronous coroutine to fetch the LookupOverrideInstance and return response metadata


        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._fetch_async()
        instance = LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def _update(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> tuple:
        """
        Internal helper for update operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = overrides_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return self._version.update_with_response_info(
            method="PUT", uri=self._uri, data=data, headers=headers
        )

    def update(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> LookupOverrideInstance:
        """
        Update the LookupOverrideInstance

        :param overrides_request:

        :returns: The updated LookupOverrideInstance
        """
        payload, _, _ = self._update(overrides_request=overrides_request)
        return LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )

    def update_with_http_info(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Update the LookupOverrideInstance and return response metadata

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._update(
            overrides_request=overrides_request
        )
        instance = LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _update_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> tuple:
        """
        Internal async helper for update operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = overrides_request.to_dict()

        headers = values.of({})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return await self._version.update_with_response_info_async(
            method="PUT", uri=self._uri, data=data, headers=headers
        )

    async def update_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> LookupOverrideInstance:
        """
        Asynchronous coroutine to update the LookupOverrideInstance

        :param overrides_request:

        :returns: The updated LookupOverrideInstance
        """
        payload, _, _ = await self._update_async(overrides_request=overrides_request)
        return LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )

    async def update_with_http_info_async(
        self, overrides_request: Union[OverridesRequest, object] = values.unset
    ) -> ApiResponse:
        """
        Asynchronous coroutine to update the LookupOverrideInstance and return response metadata

        :param overrides_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._update_async(
            overrides_request=overrides_request
        )
        instance = LookupOverrideInstance(
            self._version,
            payload,
            field=self._solution["field"],
            phone_number=self._solution["phone_number"],
        )
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Lookups.V2.LookupOverrideContext {}>".format(context)


class LookupOverrideList(ListResource):

    class OverridesRequest(object):
        """
        :ivar line_type: The new line type to override the original line type
        :ivar reason: The reason for the override
        """

        def __init__(self, payload: Dict[str, Any]):

            self.line_type: Optional["LookupOverrideInstance.str"] = payload.get(
                "line_type"
            )
            self.reason: Optional[str] = payload.get("reason")

        def to_dict(self):
            return {
                "line_type": self.line_type,
                "reason": self.reason,
            }

    def __init__(self, version: Version):
        """
        Initialize the LookupOverrideList

        :param version: Version that contains the resource

        """
        super().__init__(version)

    def get(self, field: str, phone_number: str) -> LookupOverrideContext:
        """
        Constructs a LookupOverrideContext

        :param field:
        :param phone_number:
        """
        return LookupOverrideContext(
            self._version, field=field, phone_number=phone_number
        )

    def __call__(self, field: str, phone_number: str) -> LookupOverrideContext:
        """
        Constructs a LookupOverrideContext

        :param field:
        :param phone_number:
        """
        return LookupOverrideContext(
            self._version, field=field, phone_number=phone_number
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Lookups.V2.LookupOverrideList>"



import json
from typing import Any, AsyncIterator, Dict, Iterator, Optional, Tuple, Union

from twilio.base import values
from twilio.base.domain import Domain
from twilio.base.exceptions import TwilioRestException, TwilioServiceException
from twilio.base.page import Page
from twilio.http.response import Response


class Version(object):
    """
    Represents an API version.
    """

    def __init__(self, domain: Domain, version: str):
        self.domain = domain
        self.version = version

    def absolute_url(self, uri: str) -> str:
        """
        Turns a relative uri into an absolute url.
        """
        return self.domain.absolute_url(self.relative_uri(uri))

    def relative_uri(self, uri: str) -> str:
        """
        Turns a relative uri into a versioned relative uri.
        """
        return "{}/{}".format(self.version.strip("/"), uri.strip("/"))

    def request(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Response:
        """
        Make an HTTP request.
        """
        url = self.relative_uri(uri)
        return self.domain.request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

    async def request_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Response:
        """
        Make an asynchronous HTTP request
        """
        url = self.relative_uri(uri)
        return await self.domain.request_async(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

    @classmethod
    def exception(
        cls, method: str, uri: str, response: Response, message: str
    ) -> Union[TwilioRestException, TwilioServiceException]:
        """
        Wraps an exceptional response in a `TwilioRestException` or `TwilioServiceException`.

        If the response is RFC-9457 compliant (contains 'type', 'title', 'status', and 'code' fields),
        returns a TwilioServiceException. Otherwise, returns a TwilioRestException for backward compatibility.
        """
        # noinspection PyBroadException
        try:
            error_payload = json.loads(response.text)

            # Check if this is an RFC-9457 compliant error response
            # Required fields: type, title, status, code
            if all(key in error_payload for key in ["type", "title", "status", "code"]):
                # This is an RFC-9457 compliant error response
                return TwilioServiceException(
                    type_uri=error_payload["type"],
                    title=error_payload["title"],
                    status=error_payload["status"],
                    code=error_payload["code"],
                    detail=error_payload.get("detail"),
                    instance=error_payload.get("instance"),
                    errors=error_payload.get("errors"),
                    method=method,
                    uri=uri,
                )
            else:
                # Legacy error format - use TwilioRestException
                if "message" in error_payload:
                    message = "{}: {}".format(message, error_payload["message"])
                details = error_payload.get("details")
                code = error_payload.get("code", response.status_code)
                return TwilioRestException(
                    response.status_code, uri, message, code, method, details
                )
        except Exception:
            return TwilioRestException(
                response.status_code, uri, message, response.status_code, method
            )

    def _parse_fetch(self, method: str, uri: str, response: Response) -> Any:
        """
        Parses fetch response JSON
        """
        # Note that 3XX response codes are allowed for fetches.
        if response.status_code < 200 or response.status_code >= 400:
            raise self.exception(method, uri, response, "Unable to fetch record")

        return json.loads(response.text)

    def fetch(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Any:
        """
        Fetch a resource instance.
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

        return self._parse_fetch(method, uri, response)

    async def fetch_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Any:
        """
        Asynchronously fetch a resource instance.
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        return self._parse_fetch(method, uri, response)

    def fetch_with_response_info(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Any, int, Dict[str, str]]:
        """
        Fetch a resource and return response metadata

        Returns:
            tuple: (payload_dict, status_code, headers_dict)
                - payload_dict: The JSON response body as a dictionary
                - status_code: HTTP status code (typically 200)
                - headers_dict: Response headers as a dictionary
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        payload = self._parse_fetch(method, uri, response)
        return payload, response.status_code, dict(response.headers or {})

    async def fetch_with_response_info_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Any, int, Dict[str, str]]:
        """
        Asynchronously fetch a resource and return response metadata

        Returns:
            tuple: (payload_dict, status_code, headers_dict)
                - payload_dict: The JSON response body as a dictionary
                - status_code: HTTP status code (typically 200)
                - headers_dict: Response headers as a dictionary
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        payload = self._parse_fetch(method, uri, response)
        return payload, response.status_code, dict(response.headers or {})

    def _parse_update(self, method: str, uri: str, response: Response) -> Any:
        """
        Parses update response JSON
        """
        if response.status_code < 200 or response.status_code >= 300:
            raise self.exception(method, uri, response, "Unable to update record")

        return json.loads(response.text)

    def update(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Any:
        """
        Update a resource instance.
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

        return self._parse_update(method, uri, response)

    async def update_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Any:
        """
        Asynchronously update a resource instance.
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

        return self._parse_update(method, uri, response)

    def update_with_response_info(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Any, int, Dict[str, str]]:
        """
        Update a resource and return response metadata

        Returns:
            tuple: (payload_dict, status_code, headers_dict)
                - payload_dict: The JSON response body as a dictionary
                - status_code: HTTP status code (typically 200)
                - headers_dict: Response headers as a dictionary
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        payload = self._parse_update(method, uri, response)
        return payload, response.status_code, dict(response.headers or {})

    async def update_with_response_info_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Any, int, Dict[str, str]]:
        """
        Asynchronously update a resource and return response metadata

        Returns:
            tuple: (payload_dict, status_code, headers_dict)
                - payload_dict: The JSON response body as a dictionary
                - status_code: HTTP status code (typically 200)
                - headers_dict: Response headers as a dictionary
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        payload = self._parse_update(method, uri, response)
        return payload, response.status_code, dict(response.headers or {})

    def _parse_delete(self, method: str, uri: str, response: Response) -> bool:
        """
        Parses delete response JSON
        """
        if response.status_code < 200 or response.status_code >= 300:
            raise self.exception(method, uri, response, "Unable to delete record")

        return True  # if response code is 2XX, deletion was successful

    def delete(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> bool:
        """
        Delete a resource.
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

        return self._parse_delete(method, uri, response)

    async def delete_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> bool:
        """
        Asynchronously delete a resource.
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

        return self._parse_delete(method, uri, response)

    def delete_with_response_info(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[bool, int, Dict[str, str]]:
        """
        Delete a resource and return response metadata

        Returns:
            tuple: (success_boolean, status_code, headers_dict)
                - success_boolean: True if deletion was successful (2XX response)
                - status_code: HTTP status code (typically 204 for successful delete)
                - headers_dict: Response headers as a dictionary
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        success = self._parse_delete(method, uri, response)
        return success, response.status_code, dict(response.headers or {})

    async def delete_with_response_info_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[bool, int, Dict[str, str]]:
        """
        Asynchronously delete a resource and return response metadata

        Returns:
            tuple: (success_boolean, status_code, headers_dict)
                - success_boolean: True if deletion was successful (2XX response)
                - status_code: HTTP status code (typically 204 for successful delete)
                - headers_dict: Response headers as a dictionary
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        success = self._parse_delete(method, uri, response)
        return success, response.status_code, dict(response.headers or {})

    def read_limits(
        self, limit: Optional[int] = None, page_size: Optional[int] = None
    ) -> Dict[str, object]:
        """
        Takes a limit on the max number of records to read and a max page_size
        and calculates the max number of pages to read.

        :param limit: Max number of records to read.
        :param page_size: Max page size.
        :return A dictionary of paging limits.
        """
        if limit is not None and page_size is None:
            page_size = limit

        return {
            "limit": limit or values.unset,
            "page_size": page_size or values.unset,
        }

    def page(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Response:
        """
        Makes an HTTP request.
        """
        return self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

    async def page_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Response:
        """
        Makes an asynchronous HTTP request.
        """
        return await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )

    def page_with_response_info(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Response, int, Dict[str, str]]:
        """
        Fetch a page and return response metadata

        Returns:
            tuple: (response_object, status_code, headers_dict)
                - response_object: The Response object (not parsed JSON)
                - status_code: HTTP status code
                - headers_dict: Response headers as a dictionary
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        return response, response.status_code, dict(response.headers or {})

    async def page_with_response_info_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Response, int, Dict[str, str]]:
        """
        Asynchronously fetch a page and return response metadata

        Returns:
            tuple: (response_object, status_code, headers_dict)
                - response_object: The Response object (not parsed JSON)
                - status_code: HTTP status code
                - headers_dict: Response headers as a dictionary
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        return response, response.status_code, dict(response.headers or {})

    def stream(
        self,
        page: Optional[Page],
        limit: Optional[int] = None,
        page_limit: Optional[int] = None,
    ) -> Iterator[Any]:
        """
        Generates records one a time from a page, stopping at prescribed limits.

        :param page: The page to stream.
        :param limit: The max number of records to read.
        :param page_limit: The max number of pages to read.
        """
        current_record = 1
        current_page = 1

        while page is not None:
            for record in page:
                yield record
                current_record += 1
                if limit and limit is not values.unset and limit < current_record:
                    return

            current_page += 1
            if (
                page_limit
                and page_limit is not values.unset
                and page_limit < current_page
            ):
                return

            page = page.next_page()

    async def stream_async(
        self,
        page: Optional[Page],
        limit: Optional[int] = None,
        page_limit: Optional[int] = None,
    ) -> AsyncIterator[Any]:
        """
        Generates records one a time from a page, stopping at prescribed limits.

        :param page: The page to stream.
        :param limit: The max number of records to read.
        :param page_limit: The max number of pages to read.
        """
        current_record = 1
        current_page = 1

        while page is not None:
            for record in page:
                yield record
                current_record += 1
                if limit and limit is not values.unset and limit < current_record:
                    return

            current_page += 1
            if (
                page_limit
                and page_limit is not values.unset
                and page_limit < current_page
            ):
                return

            page = await page.next_page_async()

    def _parse_create(self, method: str, uri: str, response: Response) -> Any:
        """
        Parse create response JSON
        """
        if response.status_code < 200 or response.status_code >= 300:
            raise self.exception(method, uri, response, "Unable to create record")

        return json.loads(response.text)

    def create(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Any:
        """
        Create a resource instance.
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        return self._parse_create(method, uri, response)

    async def create_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Any:
        """
        Asynchronously create a resource instance.
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        return self._parse_create(method, uri, response)

    def create_with_response_info(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Any, int, Dict[str, str]]:
        """
        Create a resource and return response metadata

        Returns:
            tuple: (payload_dict, status_code, headers_dict)
                - payload_dict: The JSON response body as a dictionary
                - status_code: HTTP status code (e.g., 201)
                - headers_dict: Response headers as a dictionary
        """
        response = self.request(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        payload = self._parse_create(method, uri, response)
        return payload, response.status_code, dict(response.headers or {})

    async def create_with_response_info_async(
        self,
        method: str,
        uri: str,
        params: Optional[Dict[str, object]] = None,
        data: Optional[Dict[str, object]] = None,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[float] = None,
        allow_redirects: bool = False,
    ) -> Tuple[Any, int, Dict[str, str]]:
        """
        Asynchronously create a resource and return response metadata

        Returns:
            tuple: (payload_dict, status_code, headers_dict)
                - payload_dict: The JSON response body as a dictionary
                - status_code: HTTP status code (e.g., 201)
                - headers_dict: Response headers as a dictionary
        """
        response = await self.request_async(
            method,
            uri,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
        )
        payload = self._parse_create(method, uri, response)
        return payload, response.status_code, dict(response.headers or {})



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Assistants
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, values
from twilio.base.api_response import ApiResponse

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class PolicyInstance(InstanceResource):
    """
    :ivar id: The Policy ID.
    :ivar name: The name of the policy.
    :ivar description: The description of the policy.
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Policy resource.
    :ivar user_sid: The SID of the User that created the Policy resource.
    :ivar type: The type of the policy.
    :ivar policy_details: The details of the policy based on the type.
    :ivar date_created: The date and time in GMT when the Policy was created specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar date_updated: The date and time in GMT when the Policy was last updated specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    """

    def __init__(self, version: Version, payload: Dict[str, Any]):
        super().__init__(version)

        self.id: Optional[str] = payload.get("id")
        self.name: Optional[str] = payload.get("name")
        self.description: Optional[str] = payload.get("description")
        self.account_sid: Optional[str] = payload.get("account_sid")
        self.user_sid: Optional[str] = payload.get("user_sid")
        self.type: Optional[str] = payload.get("type")
        self.policy_details: Optional[Dict[str, object]] = payload.get("policy_details")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.date_updated: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_updated")
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """

        return "<Twilio.Assistants.V1.PolicyInstance>"


class PolicyPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> PolicyInstance:
        """
        Build an instance of PolicyInstance

        :param payload: Payload response from the API
        """
        return PolicyInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Assistants.V1.PolicyPage>"


class PolicyList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the PolicyList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/Policies"

    def stream(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[PolicyInstance]:
        """
        Streams PolicyInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(
            tool_id=tool_id, knowledge_id=knowledge_id, page_size=limits["page_size"]
        )

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[PolicyInstance]:
        """
        Asynchronously streams PolicyInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            tool_id=tool_id, knowledge_id=knowledge_id, page_size=limits["page_size"]
        )

        return self._version.stream_async(page, limits["limit"])

    def stream_with_http_info(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> tuple:
        """
        Streams PolicyInstance and returns headers from first page


        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: tuple of (generator, status_code, headers) where generator yields instances
        """
        limits = self._version.read_limits(limit, page_size)
        page_response = self.page_with_http_info(
            tool_id=tool_id, knowledge_id=knowledge_id, page_size=limits["page_size"]
        )

        generator = self._version.stream(page_response.data, limits["limit"])
        return (generator, page_response.status_code, page_response.headers)

    async def stream_with_http_info_async(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> tuple:
        """
        Asynchronously streams PolicyInstance and returns headers from first page


        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: tuple of (generator, status_code, headers) where generator yields instances
        """
        limits = self._version.read_limits(limit, page_size)
        page_response = await self.page_with_http_info_async(
            tool_id=tool_id, knowledge_id=knowledge_id, page_size=limits["page_size"]
        )

        generator = self._version.stream_async(page_response.data, limits["limit"])
        return (generator, page_response.status_code, page_response.headers)

    def list(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[PolicyInstance]:
        """
        Lists PolicyInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                tool_id=tool_id,
                knowledge_id=knowledge_id,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[PolicyInstance]:
        """
        Asynchronously lists PolicyInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                tool_id=tool_id,
                knowledge_id=knowledge_id,
                limit=limit,
                page_size=page_size,
            )
        ]

    def list_with_http_info(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ApiResponse:
        """
        Lists PolicyInstance and returns headers from first page


        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: ApiResponse with list of instances, status code, and headers
        """
        generator, status_code, headers = self.stream_with_http_info(
            tool_id=tool_id,
            knowledge_id=knowledge_id,
            limit=limit,
            page_size=page_size,
        )
        items = list(generator)
        return ApiResponse(data=items, status_code=status_code, headers=headers)

    async def list_with_http_info_async(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ApiResponse:
        """
        Asynchronously lists PolicyInstance and returns headers from first page


        :param str tool_id: The tool ID.
        :param str knowledge_id: The knowledge ID.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: ApiResponse with list of instances, status code, and headers
        """
        generator, status_code, headers = await self.stream_with_http_info_async(
            tool_id=tool_id,
            knowledge_id=knowledge_id,
            limit=limit,
            page_size=page_size,
        )
        items = [record async for record in generator]
        return ApiResponse(data=items, status_code=status_code, headers=headers)

    def page(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> PolicyPage:
        """
        Retrieve a single page of PolicyInstance records from the API.
        Request is executed immediately

        :param tool_id: The tool ID.
        :param knowledge_id: The knowledge ID.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of PolicyInstance
        """
        data = values.of(
            {
                "ToolId": tool_id,
                "KnowledgeId": knowledge_id,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return PolicyPage(self._version, response)

    async def page_async(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> PolicyPage:
        """
        Asynchronously retrieve a single page of PolicyInstance records from the API.
        Request is executed immediately

        :param tool_id: The tool ID.
        :param knowledge_id: The knowledge ID.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of PolicyInstance
        """
        data = values.of(
            {
                "ToolId": tool_id,
                "KnowledgeId": knowledge_id,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return PolicyPage(self._version, response)

    def page_with_http_info(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> ApiResponse:
        """
        Retrieve a single page with response metadata


        :param tool_id: The tool ID.
        :param knowledge_id: The knowledge ID.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: ApiResponse with PolicyPage, status code, and headers
        """
        data = values.of(
            {
                "ToolId": tool_id,
                "KnowledgeId": knowledge_id,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response, status_code, response_headers = self._version.page_with_response_info(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        page = PolicyPage(self._version, response)
        return ApiResponse(data=page, status_code=status_code, headers=response_headers)

    async def page_with_http_info_async(
        self,
        tool_id: Union[str, object] = values.unset,
        knowledge_id: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> ApiResponse:
        """
        Asynchronously retrieve a single page with response metadata


        :param tool_id: The tool ID.
        :param knowledge_id: The knowledge ID.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: ApiResponse with PolicyPage, status code, and headers
        """
        data = values.of(
            {
                "ToolId": tool_id,
                "KnowledgeId": knowledge_id,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response, status_code, response_headers = (
            await self._version.page_with_response_info_async(
                method="GET", uri=self._uri, params=data, headers=headers
            )
        )
        page = PolicyPage(self._version, response)
        return ApiResponse(data=page, status_code=status_code, headers=response_headers)

    def get_page(self, target_url: str) -> PolicyPage:
        """
        Retrieve a specific page of PolicyInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of PolicyInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return PolicyPage(self._version, response)

    async def get_page_async(self, target_url: str) -> PolicyPage:
        """
        Asynchronously retrieve a specific page of PolicyInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of PolicyInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return PolicyPage(self._version, response)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Assistants.V1.PolicyList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Assistants
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, values
from twilio.base.api_response import ApiResponse

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class FeedbackInstance(InstanceResource):

    class AssistantsV1ServiceCreateFeedbackRequest(object):
        """
        :ivar message_id: The message ID.
        :ivar score: The score to be given(0-1).
        :ivar session_id: The Session ID.
        :ivar text: The text to be given as feedback.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.message_id: Optional[str] = payload.get("message_id")
            self.score: Optional[float] = payload.get("score")
            self.session_id: Optional[str] = payload.get("session_id")
            self.text: Optional[str] = payload.get("text")

        def to_dict(self):
            return {
                "message_id": self.message_id,
                "score": self.score,
                "session_id": self.session_id,
                "text": self.text,
            }

    """
    :ivar assistant_id: The Assistant ID.
    :ivar id: The Feedback ID.
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Feedback.
    :ivar user_sid: The SID of the User created the Feedback.
    :ivar message_id: The Message ID.
    :ivar score: The Score to provide as Feedback (0-1)
    :ivar session_id: The Session ID.
    :ivar text: The text to be given as feedback.
    :ivar date_created: The date and time in GMT when the Feedback was created specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar date_updated: The date and time in GMT when the Feedback was last updated specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    """

    def __init__(self, version: Version, payload: Dict[str, Any], id: str):
        super().__init__(version)

        self.assistant_id: Optional[str] = payload.get("assistant_id")
        self.id: Optional[str] = payload.get("id")
        self.account_sid: Optional[str] = payload.get("account_sid")
        self.user_sid: Optional[str] = payload.get("user_sid")
        self.message_id: Optional[str] = payload.get("message_id")
        self.score: Optional[float] = payload.get("score")
        self.session_id: Optional[str] = payload.get("session_id")
        self.text: Optional[str] = payload.get("text")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.date_updated: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_updated")
        )

        self._solution = {
            "id": id,
        }

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Assistants.V1.FeedbackInstance {}>".format(context)


class FeedbackPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> FeedbackInstance:
        """
        Build an instance of FeedbackInstance

        :param payload: Payload response from the API
        """
        return FeedbackInstance(self._version, payload, id=self._solution["id"])

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Assistants.V1.FeedbackPage>"


class FeedbackList(ListResource):

    class AssistantsV1ServiceCreateFeedbackRequest(object):
        """
        :ivar message_id: The message ID.
        :ivar score: The score to be given(0-1).
        :ivar session_id: The Session ID.
        :ivar text: The text to be given as feedback.
        """

        def __init__(self, payload: Dict[str, Any]):

            self.message_id: Optional[str] = payload.get("message_id")
            self.score: Optional[float] = payload.get("score")
            self.session_id: Optional[str] = payload.get("session_id")
            self.text: Optional[str] = payload.get("text")

        def to_dict(self):
            return {
                "message_id": self.message_id,
                "score": self.score,
                "session_id": self.session_id,
                "text": self.text,
            }

    def __init__(self, version: Version, id: str):
        """
        Initialize the FeedbackList

        :param version: Version that contains the resource
        :param id: The assistant ID.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "id": id,
        }
        self._uri = "/Assistants/{id}/Feedbacks".format(**self._solution)

    def _create(
        self,
        assistants_v1_service_create_feedback_request: AssistantsV1ServiceCreateFeedbackRequest,
    ) -> tuple:
        """
        Internal helper for create operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = assistants_v1_service_create_feedback_request.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return self._version.create_with_response_info(
            method="POST", uri=self._uri, data=data, headers=headers
        )

    def create(
        self,
        assistants_v1_service_create_feedback_request: AssistantsV1ServiceCreateFeedbackRequest,
    ) -> FeedbackInstance:
        """
        Create the FeedbackInstance

        :param assistants_v1_service_create_feedback_request:

        :returns: The created FeedbackInstance
        """
        payload, _, _ = self._create(
            assistants_v1_service_create_feedback_request=assistants_v1_service_create_feedback_request
        )
        return FeedbackInstance(self._version, payload, id=self._solution["id"])

    def create_with_http_info(
        self,
        assistants_v1_service_create_feedback_request: AssistantsV1ServiceCreateFeedbackRequest,
    ) -> ApiResponse:
        """
        Create the FeedbackInstance and return response metadata

        :param assistants_v1_service_create_feedback_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = self._create(
            assistants_v1_service_create_feedback_request=assistants_v1_service_create_feedback_request
        )
        instance = FeedbackInstance(self._version, payload, id=self._solution["id"])
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    async def _create_async(
        self,
        assistants_v1_service_create_feedback_request: AssistantsV1ServiceCreateFeedbackRequest,
    ) -> tuple:
        """
        Internal async helper for create operation

        Returns:
            tuple: (payload, status_code, headers)
        """
        data = assistants_v1_service_create_feedback_request.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        return await self._version.create_with_response_info_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

    async def create_async(
        self,
        assistants_v1_service_create_feedback_request: AssistantsV1ServiceCreateFeedbackRequest,
    ) -> FeedbackInstance:
        """
        Asynchronously create the FeedbackInstance

        :param assistants_v1_service_create_feedback_request:

        :returns: The created FeedbackInstance
        """
        payload, _, _ = await self._create_async(
            assistants_v1_service_create_feedback_request=assistants_v1_service_create_feedback_request
        )
        return FeedbackInstance(self._version, payload, id=self._solution["id"])

    async def create_with_http_info_async(
        self,
        assistants_v1_service_create_feedback_request: AssistantsV1ServiceCreateFeedbackRequest,
    ) -> ApiResponse:
        """
        Asynchronously create the FeedbackInstance and return response metadata

        :param assistants_v1_service_create_feedback_request:

        :returns: ApiResponse with instance, status code, and headers
        """
        payload, status_code, headers = await self._create_async(
            assistants_v1_service_create_feedback_request=assistants_v1_service_create_feedback_request
        )
        instance = FeedbackInstance(self._version, payload, id=self._solution["id"])
        return ApiResponse(data=instance, status_code=status_code, headers=headers)

    def stream(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[FeedbackInstance]:
        """
        Streams FeedbackInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[FeedbackInstance]:
        """
        Asynchronously streams FeedbackInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(page_size=limits["page_size"])

        return self._version.stream_async(page, limits["limit"])

    def stream_with_http_info(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> tuple:
        """
        Streams FeedbackInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: tuple of (generator, status_code, headers) where generator yields instances
        """
        limits = self._version.read_limits(limit, page_size)
        page_response = self.page_with_http_info(page_size=limits["page_size"])

        generator = self._version.stream(page_response.data, limits["limit"])
        return (generator, page_response.status_code, page_response.headers)

    async def stream_with_http_info_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> tuple:
        """
        Asynchronously streams FeedbackInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: tuple of (generator, status_code, headers) where generator yields instances
        """
        limits = self._version.read_limits(limit, page_size)
        page_response = await self.page_with_http_info_async(
            page_size=limits["page_size"]
        )

        generator = self._version.stream_async(page_response.data, limits["limit"])
        return (generator, page_response.status_code, page_response.headers)

    def list(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[FeedbackInstance]:
        """
        Lists FeedbackInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[FeedbackInstance]:
        """
        Asynchronously lists FeedbackInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                limit=limit,
                page_size=page_size,
            )
        ]

    def list_with_http_info(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ApiResponse:
        """
        Lists FeedbackInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: ApiResponse with list of instances, status code, and headers
        """
        generator, status_code, headers = self.stream_with_http_info(
            limit=limit,
            page_size=page_size,
        )
        items = list(generator)
        return ApiResponse(data=items, status_code=status_code, headers=headers)

    async def list_with_http_info_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ApiResponse:
        """
        Asynchronously lists FeedbackInstance and returns headers from first page


        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: ApiResponse with list of instances, status code, and headers
        """
        generator, status_code, headers = await self.stream_with_http_info_async(
            limit=limit,
            page_size=page_size,
        )
        items = [record async for record in generator]
        return ApiResponse(data=items, status_code=status_code, headers=headers)

    def page(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> FeedbackPage:
        """
        Retrieve a single page of FeedbackInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of FeedbackInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return FeedbackPage(self._version, response, self._solution)

    async def page_async(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> FeedbackPage:
        """
        Asynchronously retrieve a single page of FeedbackInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of FeedbackInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return FeedbackPage(self._version, response, self._solution)

    def page_with_http_info(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> ApiResponse:
        """
        Retrieve a single page with response metadata


        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: ApiResponse with FeedbackPage, status code, and headers
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response, status_code, response_headers = self._version.page_with_response_info(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        page = FeedbackPage(self._version, response, self._solution)
        return ApiResponse(data=page, status_code=status_code, headers=response_headers)

    async def page_with_http_info_async(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> ApiResponse:
        """
        Asynchronously retrieve a single page with response metadata


        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: ApiResponse with FeedbackPage, status code, and headers
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response, status_code, response_headers = (
            await self._version.page_with_response_info_async(
                method="GET", uri=self._uri, params=data, headers=headers
            )
        )
        page = FeedbackPage(self._version, response, self._solution)
        return ApiResponse(data=page, status_code=status_code, headers=response_headers)

    def get_page(self, target_url: str) -> FeedbackPage:
        """
        Retrieve a specific page of FeedbackInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of FeedbackInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return FeedbackPage(self._version, response, self._solution)

    async def get_page_async(self, target_url: str) -> FeedbackPage:
        """
        Asynchronously retrieve a specific page of FeedbackInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of FeedbackInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return FeedbackPage(self._version, response, self._solution)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Assistants.V1.FeedbackList>"
